/*----------------------------------------------------------------------*/
/*		msort.cpp		test sort										*/
/*																		*/
/*		Jeff Reid		07SEP15  15:00									*/
/*----------------------------------------------------------------------*/
/*		abbreviations for name prefixes 								*/
/*																		*/
/*		a		array of ...											*/
/*		b		unsigned char (byte)									*/
/*		cb		count of bytes in ...									*/
/*		fp		FILE pointer											*/
/*		p		pointer to ...											*/
/*		dw		unsigned int (dword)									*/
/*----------------------------------------------------------------------*/

/*		get rid of warnings for fopen */
#define _CRT_SECURE_NO_WARNINGS

#include <algorithm>
#include <fstream>
#include <iomanip>
#include <iostream>
#include <vector>

#include <stdio.h>
#include <stdlib.h>
#include <time.h>

#define NUMREC (1024*4096)				// number of records

typedef unsigned __int64	UI64;
typedef unsigned __int64 *	PUI64;

#define CMPFUNC 0						// set to 1 to use compare function

#if CMPFUNC
typedef bool (*PFCMP)(UI64, UI64);
static bool ElemCmp(UI64, UI64);
PFCMP pfCmp = ElemCmp;
static UI64 qwCmp;
#endif

static clock_t dwTimeStart;				// clock values
static clock_t dwTimeStop;

static PUI64	pSort;					// ptr to sorted data
static FILE *	fpDst;					// file desc dest file

static void MakeGroups(PUI64);
static void MergeGroups(PUI64, PUI64);

int main( )
{
using namespace std;

	vector <UI64> vData (NUMREC);
	vector <UI64> vTemp (NUMREC);
	UI64 i, j;

//		generate data

	for(j = 0; j < NUMREC; j++){
		i  = (((UI64)((rand()>>4) & 0xff))<< 0);
		i += (((UI64)((rand()>>4) & 0xff))<< 8);
		i += (((UI64)((rand()>>4) & 0xff))<<16);
		i += (((UI64)((rand()>>4) & 0xff))<<24);
		i += (((UI64)((rand()>>4) & 0xff))<<32);
		i += (((UI64)((rand()>>4) & 0xff))<<40);
		i += (((UI64)((rand()>>4) & 0xff))<<48);
		i += (((UI64)((rand()>>4) & 0xff))<<56);
		*(((PUI64)&vData[0])+j) = i ;}

//		sort data

	dwTimeStart = clock();

	MakeGroups(&vData[0]);

	MergeGroups(&vData[0], &vTemp[0]);

	dwTimeStop = clock();
	cout << "Number of ticks    " << (dwTimeStop-dwTimeStart) << "\n";

#if CMPFUNC
	cout << "Number of compares " << qwCmp << "\n";
#endif

//		output data

	fpDst = fopen("dstm.bin", "wb");
	if(fpDst == NULL){
		goto exit0;}
	fwrite(pSort, 1, NUMREC*sizeof(UI64), fpDst);
	fclose(fpDst);

exit0:
	return(0);
}

#if CMPFUNC
/*----------------------------------------------------------------------*/
/*		ElemCmp		element compare										*/
/*----------------------------------------------------------------------*/
static bool ElemCmp(UI64 elem1, UI64 elem2)
{
	qwCmp += 1;							/* bump # of compares */
	return elem1 <= elem2;				/* return compare value */
}
#endif


/*----------------------------------------------------------------------*/
/*		MakeGroups														*/
/*----------------------------------------------------------------------*/
static void MakeGroups(PUI64 pv)
{
PUI64	pe;
UI64	qw0, qw1;

	pe = pv + (NUMREC&0xfffffffeul);
	while(pv < pe){
		qw0 = pv[0];
		qw1 = pv[1];
#if CMPFUNC
		if(!pfCmp(qw0, qw1)){
#else
		if(qw0 > qw1){
#endif
			pv[0] = qw1;
			pv[1] = qw0;}
		pv += 2;}
}


/*----------------------------------------------------------------------*/
/*		MergeGroups														*/
/*----------------------------------------------------------------------*/
static void MergeGroups(PUI64 pv0, PUI64 pv1)
{
PUI64	pGet;							/* current pass base ptrs */
PUI64	pPut;
PUI64	pGet0;   						/* working ptrs */
PUI64	pGet1;
PUI64	pEnd;
PUI64	pEnd0;
PUI64	pEnd1;
PUI64	pPutx;
UI64	qwRec0;							/* working records */
UI64	qwRec1;
UI64	qwGrpCnt;						/* group size */

	pGet = pv0;							/* set up for first pass */
	pPut = pv1;
	qwGrpCnt = 2;

/*		start of a pass */

	while(qwGrpCnt < NUMREC){			/* while not done */
		pGet1 = pGet;					/* update pass ptrs */
		pPutx = pPut;
		pEnd  = pGet+NUMREC;

/*		merge two groups */

		while(1){
			pGet0 = pGet1;				/* update group ptrs */
			pEnd0 = pGet0 + qwGrpCnt;
			if(pEnd0 >= pEnd)
				pEnd0 = pEnd;
		    pGet1 = pEnd0;
			pEnd1 = pGet1 + qwGrpCnt;
			if(pEnd1 >= pEnd)
				pEnd1 = pEnd;
			if(pGet0 == pEnd0)			/* if no data, end pass */
				break;
			if(pGet1 == pEnd1){			/* if only group0, copy it, end pass */
				while(pGet0 != pEnd0)
					*pPutx++ = *pGet0++;
				break;}

/*		repeatedly copy the lower of two records */
/*		until both groups copied */

			qwRec0 = *pGet0++;			/* get 1st 2 records */
			qwRec1 = *pGet1++;
			while(1){
#if CMPFUNC
				if(pfCmp(qwRec0, qwRec1)){	/* if 0 <= 1, put rec0 */
#else
				if(qwRec0 <= qwRec1){	/* if 0 <= 1, put rec0 */
#endif
					*pPutx++ = qwRec0;
					if(pGet0 < pEnd0){	/*   if not end 0, get next rec0 */
						qwRec0 = *pGet0++;
						continue;}
					else{				/*   else copy rest of grp1 */
						*pPutx++ = qwRec1;
						while(pGet1 != pEnd1)
							*pPutx++ = *pGet1++;
						break;}}		/*   end two group merge */
				else{                   /* else 1 < 0, put rec1 */
					*pPutx++ = qwRec1;
					if(pGet1 < pEnd1){	/*   if not end 1, get next rec1 */
						qwRec1 = *pGet1++;
						continue;}
					else{				/*   else copy rest of grp0 */
						*pPutx++ = qwRec0;
						while(pGet0 != pEnd0)
							*pPutx++ = *pGet0++;
						break;}}}}		/*   end two group merge */
		pEnd = pPut;					/* swap ptrs */
		pPut = pGet;
		pGet = pEnd;
		qwGrpCnt <<= 1;}				/* bump group size */

	pSort = pGet;		 				/* set ptr to sorted data */
}
